@@ -2,9 +2,15 @@
 // Copyright (C) 2010-2021 Codership Oy <info@codership.com>
 //
 
+#include "wsrep_api.h"
+#include "wsrep_ps.h"
+
 #include "key_data.hpp"
 #include "gu_serialize.hpp"
-#include "gu_asio.hpp" //  // gu::init_allowlist_service_v1()
+#include "gu_asio.hpp" // gu::init_tls_service_v1(), gu::init_allowlist_service_v1()
+#include "gu_thread_keys.hpp"
+#include "gu_asio.hpp" // gu::init_tls_service_v1()
+#include "wsrep_membership_service.h"
 
 #if defined(GALERA_MULTIMASTER)
 #include "replicator_smm.hpp"
@@ -14,9 +20,8 @@
 #endif
 
 #include "wsrep_params.hpp"
-#include "gu_event_service.hpp"
+#include "event_service.hpp"
 #include "wsrep_config_service.h"
-#include "wsrep_node_isolation.h"
 
 #include <cassert>
 
@@ -156,9 +161,28 @@ char* galera_parameters_get (wsrep_t* gh)
 }
 
 extern "C"
-wsrep_status_t galera_enc_set_key(wsrep_t* gh, const wsrep_enc_key_t*key)
+wsrep_status_t galera_enc_set_key(wsrep_t* gh, const wsrep_enc_key_t* key)
 {
-    return WSREP_NOT_IMPLEMENTED;
+    static wsrep_enc_key_t null_key = { NULL, 0 };
+
+    REPL_CLASS * repl(reinterpret_cast< REPL_CLASS * >(gh->ctx));
+
+    if (NULL == key) key = &null_key;
+
+    try
+    {
+        return repl->enc_set_key(*key);
+    }
+    catch(std::exception& e)
+    {
+        log_error << e.what();
+        return WSREP_NODE_FAIL;
+    }
+    catch(...)
+    {
+        log_fatal << "non-standard exception";
+        return WSREP_FATAL;
+    }
 }
 
 extern "C"
@@ -518,12 +542,12 @@ galera_terminate_trx(wsrep_t*           const gh,
     return retval;
 }
 
-extern "C" wsrep_status_t wsrep_certify_v1(wsrep_t* gh,
-                                           wsrep_conn_id_t conn_id,
-                                           wsrep_ws_handle_t* trx_handle,
-                                           uint32_t flags,
-                                           wsrep_trx_meta_t* meta,
-                                           const wsrep_seq_cb_t* seq_cb)
+extern "C"
+wsrep_status_t galera_certify(wsrep_t*           const gh,
+                              wsrep_conn_id_t    const conn_id,
+                              wsrep_ws_handle_t* const trx_handle,
+                              uint32_t           const flags,
+                              wsrep_trx_meta_t*  const meta)
 {
     assert(gh != 0);
     assert(gh->ctx != 0);
@@ -611,7 +635,7 @@ extern "C" wsrep_status_t wsrep_certify_v1(wsrep_t* gh,
             }
         }
 
-        retval = repl->replicate(trx, meta, seq_cb);
+        retval = repl->replicate(trx, meta);
 
         if (meta)
         {
@@ -684,16 +708,6 @@ extern "C" wsrep_status_t wsrep_certify_v1(wsrep_t* gh,
     return retval;
 }
 
-extern "C"
-wsrep_status_t galera_certify(wsrep_t*           const gh,
-                              wsrep_conn_id_t    const conn_id,
-                              wsrep_ws_handle_t* const trx_handle,
-                              uint32_t           const flags,
-                              wsrep_trx_meta_t*  const meta)
-{
-  return wsrep_certify_v1(gh, conn_id, trx_handle, flags, meta, nullptr);
-}
-
 
 extern "C"
 wsrep_status_t galera_commit_order_enter(
@@ -988,25 +1002,14 @@ wsrep_status_t galera_append_key(wsrep_t*           const gh,
 
     try
     {
-        int const proto_ver(repl->trx_proto_ver());
         TrxHandleLock lock(*trx);
-
-        if (keys_num > 0)
+        for (size_t i(0); i < keys_num; ++i)
         {
-            for (size_t i(0); i < keys_num; ++i)
-            {
-                galera::KeyData const k(proto_ver,
-                                        keys[i].key_parts,
-                                        keys[i].key_parts_num,
-                                        key_type,
-                                        copy);
-                gu_trace(trx->append_key(k));
-            }
-        }
-        else if (proto_ver >= 6)
-        {
-            /* Append server-level key (matches every trx)*/
-            galera::KeyData const k(proto_ver, key_type);
+            galera::KeyData k (repl->trx_proto_ver(),
+                               keys[i].key_parts,
+                               keys[i].key_parts_num,
+                               key_type,
+                               copy);
             gu_trace(trx->append_key(k));
         }
         retval = WSREP_OK;
@@ -1220,7 +1223,7 @@ wsrep_status_t galera_to_execute_start(wsrep_t*                const gh,
 
         if (trx.nbo_end() == false)
         {
-            retval = repl->replicate(trx, meta, nullptr);
+            retval = repl->replicate(trx, meta);
             assert((retval == WSREP_OK && trx.ts() != 0 &&
                     trx.ts()->global_seqno() > 0) ||
                    (retval != WSREP_OK && (trx.ts() == 0  ||
@@ -1487,7 +1490,7 @@ wsrep_seqno_t galera_pause (wsrep_t* gh)
     }
     catch (gu::Exception& e)
     {
-        log_warn << "Node pause failed: " << e.what();
+        log_error << e.what();
         return -e.get_errno();
     }
 }
@@ -1508,7 +1511,7 @@ wsrep_status_t galera_resume (wsrep_t* gh)
     }
     catch (gu::Exception& e)
     {
-        log_error << "Node resume failed: " << e.what();
+        log_error << e.what();
         return WSREP_NODE_FAIL;
     }
 }
@@ -1529,7 +1532,7 @@ wsrep_status_t galera_desync (wsrep_t* gh)
     }
     catch (gu::Exception& e)
     {
-        log_warn << "Node desync failed: " << e.what();
+        log_error << e.what();
         return WSREP_TRX_FAIL;
     }
 }
@@ -1550,7 +1553,7 @@ wsrep_status_t galera_resync (wsrep_t* gh)
     }
     catch (gu::Exception& e)
     {
-        log_error << "Node resync failed: " << e.what();
+        log_error << e.what();
         return WSREP_NODE_FAIL;
     }
 }
@@ -1664,6 +1667,30 @@ int wsrep_loader(wsrep_t *hptr)
     return WSREP_OK;
 }
 
+extern "C"
+int wsrep_init_thread_service_v1(wsrep_thread_service_v1_t* ts)
+
+{
+    return gu::init_thread_service_v1(ts);
+}
+
+extern "C"
+void wsrep_deinit_thread_service_v1()
+{
+    gu::deinit_thread_service_v1();
+}
+
+extern "C"
+int wsrep_init_tls_service_v1(wsrep_tls_service_v1_t *tls_service)
+{
+    return gu::init_tls_service_v1(tls_service);
+}
+
+extern "C" void wsrep_deinit_tls_service_v1()
+{
+    gu::deinit_tls_service_v1();
+}
+
 extern "C"
 int wsrep_init_allowlist_service_v1(wsrep_allowlist_service_v1_t *allowlist_service)
 {
@@ -1678,12 +1705,12 @@ extern "C" void wsrep_deinit_allowlist_service_v1()
 extern "C"
 int wsrep_init_event_service_v1(wsrep_event_service_v1_t *event_service)
 {
-    return gu::EventService::init_v1(event_service);
+    return galera::EventService::init_v1(event_service);
 }
 
 extern "C" void wsrep_deinit_event_service_v1()
 {
-    gu::EventService::deinit_v1();
+    galera::EventService::deinit_v1();
 }
 
 static int map_parameter_flags(int flags)
@@ -1699,8 +1726,6 @@ static int map_parameter_flags(int flags)
       ret |= WSREP_PARAM_TYPE_INTEGER;
     if (flags & gu::Config::Flag::type_double)
       ret |= WSREP_PARAM_TYPE_DOUBLE;
-    if (flags & gu::Config::Flag::type_duration)
-      ret |= WSREP_PARAM_TYPE_DOUBLE;
     return ret;
 }
 
@@ -1717,30 +1742,13 @@ static int wsrep_parameter_init(wsrep_parameter& wsrep_param,
         ret = gu_str2bool(param.value().c_str(), &wsrep_param.value.as_bool);
         break;
     case gu::Config::Flag::type_integer:
-    {
-        long long tmp;
-        ret = gu_str2ll(param.value().c_str(), &tmp);
-        wsrep_param.value.as_integer = tmp;
+        long long llval;
+        ret = gu_str2ll(param.value().c_str(), &llval);
+        wsrep_param.value.as_integer = llval;
         break;
-    }
     case gu::Config::Flag::type_double:
         ret = gu_str2dbl(param.value().c_str(), &wsrep_param.value.as_double);
         break;
-    case gu::Config::Flag::type_duration:
-    {
-        try
-        {
-            // durations are mapped to doubles
-            wsrep_param.value.as_double
-                = to_double(gu::datetime::Period(param.value()));
-        }
-        catch (...)
-        {
-            assert(0);
-            return 1;
-        }
-        break;
-    }
     default:
         assert((param.flags() & gu::Config::Flag::type_mask) == 0);
         wsrep_param.value.as_string = param.value().c_str();
@@ -1754,110 +1762,133 @@ static int wsrep_parameter_init(wsrep_parameter& wsrep_param,
     return 0;
 }
 
-static wsrep_status_t
-get_parameters_from_config(const gu::Config& config,
-                           wsrep_get_parameters_cb callback, void* context)
+static wsrep_status_t get_parameters(wsrep_t* gh,
+                                     wsrep_get_parameters_cb callback,
+                                     void* context)
 {
-    for (auto& i : config)
+    assert(gh != 0);
+    assert(gh->ctx != 0);
+    REPL_CLASS * repl(reinterpret_cast< REPL_CLASS * >(gh->ctx));
+    const gu::Config& config(repl->params());
+    for (auto &i : config)
     {
         const std::string& key(i.first);
         const gu::Config::Parameter& param(i.second);
         if (!param.is_hidden())
         {
             wsrep_parameter arg;
-            if (wsrep_parameter_init(arg, key, param)
-                || (callback(&arg, context) != WSREP_OK))
+            if (wsrep_parameter_init(arg, key, param) ||
+                (callback(&arg, context) != WSREP_OK))
             {
                 log_error << "Failed to initialize parameter '" << key
-                          << "', value " << param.value() << " , flags ("
-                          << gu::Config::Flag::to_string(param.flags()) << ")";
+                          << "', value " << param.value()
+                          << " , flags (" << gu::Config::Flag::to_string(param.flags())
+                          << ")";
                 return WSREP_FATAL;
             }
         }
     }
-    return WSREP_OK;
-}
 
-static wsrep_status_t
-get_parameters_v1(wsrep_t* gh, wsrep_get_parameters_cb callback, void* context)
-{
-    assert(gh != 0);
-    if (gh->ctx)
-    {
-        REPL_CLASS* repl(reinterpret_cast<REPL_CLASS*>(gh->ctx));
-        return get_parameters_from_config(repl->params(), callback, context);
-    }
-    // Attempt to call the service before galera_init()
-    return WSREP_NOT_ALLOWED;
+    return WSREP_OK;
 }
 
 extern "C"
 int wsrep_init_config_service_v1(wsrep_config_service_v1_t *config_service)
 {
-    config_service->get_parameters = get_parameters_v1;
-    // Deprecation checks will be done by application which uses
-    // the service.
-    gu::Config::disable_deprecation_check();
+    config_service->get_parameters = get_parameters;
     return WSREP_OK;
 }
 
 extern "C"
 void wsrep_deinit_config_service_v1()
 {
-    gu::Config::enable_deprecation_check();
 }
 
-
-/*
- * Version 2 allows the service to be used before provider is initialized,
- * i.e. before galera_init().
- */
-static wsrep_status_t
-get_parameters_v2(wsrep_t* gh, wsrep_get_parameters_cb callback, void* context)
+extern "C"
+wsrep_status_t
+    wsrep_ps_fetch_cluster_info_v2 (wsrep_t*            gh,
+                                    wsrep_node_info_t** nodes,
+                                    uint32_t*           size,
+                                    int32_t*            my_index,
+                                    uint32_t            max_version)
 {
     assert(gh != 0);
+    assert(gh->ctx != 0);
 
-    if (gh->ctx)
+    REPL_CLASS* repl(reinterpret_cast< REPL_CLASS * >(gh->ctx));
+
+    return repl->fetch_pfs_info(nodes, size, my_index, max_version);
+}
+
+extern "C"
+void wsrep_ps_free_cluster_info (wsrep_t* gh,
+                                 wsrep_node_info_t* nodes)
+{
+    assert(gh != 0);
+    assert(gh->ctx != 0);
+
+    REPL_CLASS* repl(reinterpret_cast< REPL_CLASS * >(gh->ctx));
+
+    repl->free_pfs_info(nodes);
+}
+
+extern "C"
+wsrep_status_t
+    wsrep_ps_fetch_node_stat_v2 (wsrep_t*            gh,
+                                 wsrep_node_stat_t** nodes,
+                                 uint32_t*           size,
+                                 int32_t*            my_index,
+                                 uint32_t            max_version)
+{
+    assert(gh != 0);
+    assert(gh->ctx != 0);
+
+    REPL_CLASS* repl(reinterpret_cast< REPL_CLASS * >(gh->ctx));
+
+    return repl->fetch_pfs_stat(nodes, size, my_index, max_version);
+}
+
+extern "C"
+void wsrep_ps_free_node_stat (wsrep_t* gh,
+                              wsrep_node_stat_t* nodes)
+{
+    assert(gh != 0);
+    assert(gh->ctx != 0);
+
+    REPL_CLASS* repl(reinterpret_cast< REPL_CLASS * >(gh->ctx));
+
+    repl->free_pfs_stat(nodes);
+}
+
+static wsrep_status_t
+get_membership(wsrep_t* const            gh,
+               wsrep_allocator_cb const  alloc,
+               struct wsrep_membership** memb)
+{
+    assert(gh != 0);
+    assert(gh->ctx != 0);
+
+    REPL_CLASS * repl(reinterpret_cast< REPL_CLASS * >(gh->ctx));
+
+    try
     {
-        REPL_CLASS* repl(reinterpret_cast<REPL_CLASS*>(gh->ctx));
-        return get_parameters_from_config(repl->params(), callback, context);
+        return repl->get_membership(alloc, memb);
     }
-    else
+    catch (std::exception& e)
     {
-        gu::Config config;
-        galera::ReplicatorSMM::InitConfig init(config, NULL, NULL);
-        return get_parameters_from_config(config, callback, context);
+        log_error << e.what();
+        return WSREP_NODE_FAIL;
     }
 }
 
-extern "C" int
-wsrep_init_config_service_v2(wsrep_config_service_v2_t* config_service)
+extern "C"
+wsrep_status_t wsrep_init_membership_service_v1(
+    struct wsrep_membership_service_v1 *membership_service)
 {
-    config_service->get_parameters = get_parameters_v2;
-    // Deprecation checks will be done by application which uses
-    // the service.
-    gu::Config::disable_deprecation_check();
+    membership_service->get_membership = get_membership;
     return WSREP_OK;
 }
 
-extern "C" void wsrep_deinit_config_service_v2()
+extern "C" void wsrep_deinit_membership_service_v1()
 {
-    gu::Config::enable_deprecation_check();
-}
-
-/*
- * This function may be called from signal handler, so make sure that
- * only 'safe' system calls and library functions are used. See
- * https://pubs.opengroup.org/onlinepubs/009695399/functions/xsh_chap02_04.html
- */
-extern "C" enum wsrep_node_isolation_result
-wsrep_node_isolation_mode_set_v1(enum wsrep_node_isolation_mode mode)
-{
-    if (mode < WSREP_NODE_ISOLATION_NOT_ISOLATED
-        || mode > WSREP_NODE_ISOLATION_FORCE_DISCONNECT)
-    {
-        return WSREP_NODE_ISOLATION_INVALID_VALUE;
-    }
-    gu::gu_asio_node_isolation_mode = mode;
-    return WSREP_NODE_ISOLATION_SUCCESS;
 }
