@@ -1,10 +1,12 @@
 //
-// Copyright (C) 2018-2019 Codership Oy <info@codership.com>
+// Copyright (C) 2018-2022 Codership Oy <info@codership.com>
 //
 
 #include <wsrep_api.h>
 extern "C" int wsrep_loader(wsrep_t*);
 
+#include <gcache_test_encryption.hpp>
+
 #include <gu_config.hpp>
 #include <gu_mutex.hpp>
 #include <gu_cond.hpp>
@@ -58,6 +60,7 @@ static const char* Defaults[] =
 #endif
     "gcache.dir",                  ".",
     "gcache.keep_pages_size",      "0",
+    "gcache.keep_plaintext_size",  "128M", /* defaults to gcache.page_size */
     "gcache.mem_size",             "0",
     "gcache.name",                 "galera.cache",
     "gcache.page_size",            "128M",
@@ -68,6 +71,7 @@ static const char* Defaults[] =
     "gcs.fc_factor",               "1.0",
     "gcs.fc_limit",                "16",
     "gcs.fc_master_slave",         "no",
+    "gcs.fc_single_primary",       "no",
     "gcs.max_packet_size",         "64500",
     "gcs.max_throttle",            "0.25",
 #if (GU_WORDSIZE == 32)
@@ -109,11 +113,13 @@ static const char* Defaults[] =
     "socket.checksum",             "2",
     "socket.recv_buf_size",        "auto",
     "socket.send_buf_size",        "auto",
+//  "socket.dynamic",              no default,
 //  "socket.ssl",                  no default,
 //  "socket.ssl_cert",             no default,
 //  "socket.ssl_cipher",           no default,
 //  "socket.ssl_compression",      no default,
 //  "socket.ssl_key",              no default,
+//  "socket.ssl_reload"            no default,
     NULL
 };
 
@@ -127,8 +133,8 @@ fill_in_expected(DefaultsMap& map, const char* def_list[])
         std::pair<std::string, std::string> param(def_list[i], def_list[i+1]);
         DefaultsMap::iterator it(map.insert(param).first);
 
-        fail_if(it == map.end(), "Failed to insert KV pair: %s = %s",
-                param.first.c_str(), param.second.c_str());
+        ck_assert_msg(it != map.end(), "Failed to insert KV pair: %s = %s",
+                      param.first.c_str(), param.second.c_str());
     }
 }
 
@@ -146,8 +152,8 @@ fill_in_real(DefaultsMap& map, wsrep_t& provider)
                                                           kv_pairs[i].second);
         DefaultsMap::iterator it(map.insert(trimmed).first);
 
-        fail_if(it == map.end(), "Failed to insert KV pair: %s = %s",
-                trimmed.first.c_str(), trimmed.second.c_str());
+        ck_assert_msg(it != map.end(), "Failed to insert KV pair: %s = %s",
+                      trimmed.first.c_str(), trimmed.second.c_str());
     }
 }
 
@@ -167,7 +173,8 @@ struct app_ctx
     wsrep_t   provider_;
     bool      connected_;
 
-    app_ctx() : mtx_(), cond_(), provider_(), connected_(false) {}
+    app_ctx() : mtx_(0), cond_(0),
+                provider_(), connected_(false) {}
 };
 
 static enum wsrep_cb_status
@@ -222,12 +229,13 @@ recv_func(void* ctx)
     wsrep_t& provider(c->provider_);
 
     wsrep_status_t const ret(provider.recv(&provider, NULL));
-    fail_if(WSREP_OK != ret, "recv() returned %d", ret);
+    ck_assert_msg(WSREP_OK == ret, "recv() returned %d", ret);
 
     return NULL;
 }
 
-START_TEST(defaults)
+static void
+defaults(bool const enc)
 {
     DefaultsMap expected_defaults, real_defaults;
 
@@ -236,45 +244,54 @@ START_TEST(defaults)
     app_ctx ctx;
     wsrep_t& provider(ctx.provider_);
     int ret = wsrep_status_t(wsrep_loader(&provider));
-    fail_if(WSREP_OK != ret);
+    ck_assert(WSREP_OK == ret);
+
+    wsrep_encrypt_cb_t const enc_cb(enc ? gcache_test_encrypt_cb : NULL);
 
     struct wsrep_init_args init_args =
         {
-            &ctx, // void* app_ctx
+            &ctx,     // void*       app_ctx
 
             /* Configuration parameters */
-            NULL, // const char* node_name
-            NULL, // const char* node_address
-            NULL, // const char* node_incoming
-            NULL, // const char* data_dir
-            NULL, // const char* options
-            0,    // int         proto_ver
+            NULL,     // const char* node_name
+            NULL,     // const char* node_address
+            NULL,     // const char* node_incoming
+            NULL,     // const char* data_dir
+            NULL,     // const char* options
+            0,        // int         proto_ver
 
             /* Application initial state information. */
-            NULL, // const wsrep_gtid_t* state_id
-            NULL, // const wsrep_buf_t*  state
+            NULL,     // const wsrep_gtid_t*    state_id
+            NULL,     // const wsrep_buf_t*     state
 
             /* Application callbacks */
-            log_cb, // wsrep_log_cb_t         logger_cb
-            conn_cb,// wsrep_connected_cb_t   connected_cb
-            view_cb,// wsrep_view_cb_t        view_handler_cb
-            NULL,   // wsrep_sst_request_cb_t sst_request_cb
-            NULL,   // wsrep_encrypt_cb_t     encrypt_cb
+            log_cb,   // wsrep_log_cb_t         logger_cb
+            conn_cb,  // wsrep_connected_cb_t   connected_cb
+            view_cb,  // wsrep_view_cb_t        view_handler_cb
+            NULL,     // wsrep_sst_request_cb_t sst_request_cb
+            enc_cb,   // wsrep_encrypt_cb_t     encrypt_cb
 
             /* Applier callbacks */
-            NULL, // wsrep_apply_cb_t      apply_cb
-            NULL, // wsrep_unordered_cb_t  unordered_cb
+            NULL,     // wsrep_apply_cb_t       apply_cb
+            NULL,     // wsrep_unordered_cb_t   unordered_cb
 
             /* State Snapshot Transfer callbacks */
-            NULL, // wsrep_sst_donate_cb_t sst_donate_cb
-            synced_cb,// wsrep_synced_cb_t synced_cb
+            NULL,     // wsrep_sst_donate_cb_t  sst_donate_cb
+            synced_cb,// wsrep_synced_cb_t      synced_cb
         };
     ret = provider.init(&provider, &init_args);
-    fail_if(WSREP_OK != ret);
+    ck_assert(WSREP_OK == ret);
+    if (enc_cb)
+    {
+        const char* key = "oaiasjfewoer";
+        wsrep_buf buf = { key, strlen(key) };
+        ret = provider.enc_set_key(&provider, &buf);
+        ck_assert(WSREP_OK == ret);
+    }
 
     /* some defaults are set only on connection attmept */
     ret = provider.connect(&provider, "cluster_name", "gcomm://", "", false);
-    fail_if(WSREP_OK != ret, "connect() returned %d", ret);
+    ck_assert_msg(WSREP_OK == ret, "connect() returned %d", ret);
 
     fill_in_real(real_defaults, provider);
     mark_point();
@@ -283,7 +300,7 @@ START_TEST(defaults)
     {
         /* some configuration change events need to be received */
         gu_thread_t recv_thd;
-        gu_thread_create(&recv_thd, NULL, recv_func, &ctx);
+        gu_thread_create(NULL, &recv_thd,  recv_func, &ctx);
 
         mark_point();
 
@@ -297,18 +314,26 @@ START_TEST(defaults)
         mark_point();
 
         ret = provider.disconnect(&provider);
-        fail_if(WSREP_OK != ret, "disconnect() returned %d", ret);
+        ck_assert_msg(WSREP_OK == ret, "disconnect() returned %d", ret);
 
         ret = gu_thread_join(recv_thd, NULL);
-        fail_if(0 != ret, "Could not join thread: %d (%s)", ret, strerror(ret));
+        ck_assert_msg(0 == ret, "Could not join thread: %d (%s)",
+                      ret, strerror(ret));
     }
 
     provider.free(&provider);
     mark_point();
 
-    /* cleanup files */
+    /* cleanup leftover files */
     ::unlink(real_defaults.find("gcache.name")->second.c_str());
     ::unlink("grastate.dat");
+    if (enc)
+    {
+        static const char* page_file = "gcache.page.000000";
+        ck_assert_msg(0 == ::unlink(page_file),
+                      "Failed to cleanup page file '%s': %d(%s)",
+                      page_file, errno, strerror(errno));
+    }
 
     /* now compare expected and real maps */
     std::ostringstream err;
@@ -347,8 +372,19 @@ START_TEST(defaults)
         real_defaults.erase(real++);
     }
 
-    fail_if (!err.str().empty(), "Defaults discrepancies detected:\n%s",
-             err.str().c_str());
+    ck_assert_msg(err.str().empty(), "Defaults discrepancies detected:\n%s",
+                  err.str().c_str());
+}
+
+START_TEST(test_defaults)
+{
+    defaults(false);
+}
+END_TEST
+
+START_TEST(test_defaultsE)
+{
+    defaults(true);
 }
 END_TEST
 
@@ -358,7 +394,8 @@ Suite* defaults_suite()
     TCase* tc;
 
     tc = tcase_create("defaults");
-    tcase_add_test(tc, defaults);
+    tcase_add_test(tc, test_defaults);
+    tcase_add_test(tc, test_defaultsE);
     tcase_set_timeout(tc, 120);
     suite_add_tcase(s, tc);
 
