@@ -49,6 +49,10 @@
 #include <fstream>
 #include <mutex>
 
+static wsrep_tls_service_v1_t* gu_tls_service(0);
+
+static wsrep_allowlist_service_v1_t* gu_allowlist_service(0);
+
 //
 // AsioIpAddress wrapper
 //
@@ -207,17 +211,26 @@ gu::AsioErrorCode::AsioErrorCode()
     : value_()
     , category_(&gu_asio_system_category)
     , error_extra_()
+    , wsrep_category_()
+    , tls_stream_()
 { }
 
 gu::AsioErrorCode::AsioErrorCode(int err)
     : value_(err)
     , category_(&gu_asio_system_category)
     , error_extra_()
+    , wsrep_category_()
+    , tls_stream_()
 { }
 
 std::string gu::AsioErrorCode::message() const
 {
-    if (category_)
+    if (wsrep_category_ && gu_tls_service)
+    {
+        return gu_tls_service->error_message_get(
+            gu_tls_service->context, tls_stream_, value_, wsrep_category_);
+    }
+    else if (category_)
     {
         std::string ret(
             asio::error_code(value_, category_->native()).message());
@@ -333,12 +346,22 @@ bool exclude_ssl_error(const asio::error_code& ec)
 {
     switch (ERR_GET_REASON(ec.value()))
     {
+        // Short read errors seem to be generated quite frequently
+        // by SSL library because of broken connections. For Galera
+        // connections premature EOFs are not a problem because messages
+        // are framed and the protocols are fault tolerant by design.
+        // The error to suppress are:
+        // SSL_R_SHORT_READ - OpenSSL < 3.0
+        // SSL_R_UNEXPECTED_EOF_WHILE_READING - OpenSSL >= 3.0
 #ifdef SSL_R_SHORT_READ
     case SSL_R_SHORT_READ:
-        // Short read error seems to be generated quite frequently
-        // by SSL library, probably because broken connections.
         return true;
 #endif /* SSL_R_SHORT_READ */
+#ifdef SSL_R_UNEXPECTED_EOF_WHILE_READING
+    case SSL_R_UNEXPECTED_EOF_WHILE_READING:
+        // OpenSSL 3.0 and onwards.
+        return true;
+#endif /* SSL_R_UNEXPECTED_EOF_WHILE_READING */
     default:
         return false;
     }
@@ -505,15 +528,28 @@ static void init_use_ssl(gu::Config& conf)
 void gu::ssl_register_params(gu::Config& conf)
 {
     // register SSL config parameters
-    conf.add(gu::conf::use_ssl);
-    conf.add(gu::conf::ssl_cipher);
-    conf.add(gu::conf::ssl_compression);
-    conf.add(gu::conf::ssl_key);
-    conf.add(gu::conf::ssl_cert);
-    conf.add(gu::conf::ssl_ca);
-    conf.add(gu::conf::ssl_password_file);
-    conf.add(gu::conf::ssl_reload);
-    conf.add(gu::conf::socket_dynamic);
+    conf.add(gu::conf::use_ssl,
+             gu::Config::Flag::read_only |
+             gu::Config::Flag::type_bool);
+    conf.add(gu::conf::ssl_cipher,
+             gu::Config::Flag::read_only |
+             gu::Config::Flag::type_bool);
+    conf.add(gu::conf::ssl_compression,
+             gu::Config::Flag::read_only |
+             gu::Config::Flag::type_bool);
+    conf.add(gu::conf::ssl_key,
+             gu::Config::Flag::read_only);
+    conf.add(gu::conf::ssl_cert,
+             gu::Config::Flag::read_only);
+    conf.add(gu::conf::ssl_ca,
+             gu::Config::Flag::read_only);
+    conf.add(gu::conf::ssl_password_file,
+             gu::Config::Flag::read_only);
+    conf.add(gu::conf::ssl_reload,
+             gu::Config::Flag::type_bool);
+    conf.add(gu::conf::socket_dynamic,
+             gu::Config::Flag::read_only |
+             gu::Config::Flag::type_bool);
 }
 
 void gu::ssl_param_set(const std::string& key, const std::string& val, 
@@ -557,6 +593,7 @@ void gu::ssl_init_options(gu::Config& conf)
     if (use_ssl == true)
     {
         // set defaults
+        conf.set(conf::ssl_reload, 1);
 
         // cipher list
         const std::string cipher_list(conf.get(conf::ssl_cipher, ""));
@@ -632,6 +669,7 @@ bool gu::is_verbose_error(const gu::AsioErrorCode& ec)
 gu::AsioIoService::AsioIoService(const gu::Config& conf)
     : impl_(std::unique_ptr<Impl>(new Impl))
     , conf_(conf)
+    , tls_service_(gu_tls_service)
     , signal_connection_()
     , dynamic_socket_(false)
 {
@@ -695,6 +733,11 @@ void gu::AsioIoService::run_one()
     impl_->native().run_one();
 }
 
+void gu::AsioIoService::poll_one()
+{
+    impl_->native().poll_one();
+}
+
 void gu::AsioIoService::run()
 {
     impl_->native().run();
@@ -809,3 +852,80 @@ void gu::AsioSteadyTimer::cancel()
 {
     impl_->native().cancel();
 }
+
+//
+// TLS service hooks.
+//
+
+static std::mutex gu_tls_service_init_mutex;
+static size_t gu_tls_service_usage;
+
+int gu::init_tls_service_v1(wsrep_tls_service_v1_t* tls_service)
+{
+    std::lock_guard<std::mutex> lock(gu_tls_service_init_mutex);
+    ++gu_tls_service_usage;
+    if (gu_tls_service)
+    {
+        assert(gu_tls_service == tls_service);
+        return 0;
+    }
+    gu_tls_service = tls_service;
+    return 0;
+}
+
+void gu::deinit_tls_service_v1()
+{
+    std::lock_guard<std::mutex> lock(gu_tls_service_init_mutex);
+    assert(gu_tls_service_usage > 0);
+    --gu_tls_service_usage;
+    if (gu_tls_service_usage == 0) gu_tls_service = 0;
+}
+
+//
+// Allowlist
+//
+
+bool gu::allowlist_value_check(wsrep_allowlist_key_t key, const std::string& value)
+{
+    if (gu_allowlist_service == nullptr)
+    {
+        return true;
+    }
+    wsrep_buf_t const check_value = { value.c_str(), value.length() };
+    wsrep_status_t result(gu_allowlist_service->allowlist_cb(
+        gu_allowlist_service->context, key, &check_value));
+    switch (result)
+    {
+        case WSREP_OK:
+            return true;
+        case WSREP_NOT_ALLOWED:
+            return false;
+        default:
+            gu_throw_error(EINVAL) << "Unknown allowlist callback response: " << result 
+                                   << ", aborting.";
+    }
+}
+
+static std::mutex gu_allowlist_service_init_mutex;
+static size_t gu_allowlist_service_usage;
+
+int gu::init_allowlist_service_v1(wsrep_allowlist_service_v1_t* allowlist_service)
+{
+    std::lock_guard<std::mutex> lock(gu_allowlist_service_init_mutex);
+    ++gu_allowlist_service_usage;
+    if (gu_allowlist_service)
+    {
+        assert(gu_allowlist_service == allowlist_service);
+        return 0;
+    }
+    gu_allowlist_service = allowlist_service;
+    return 0;
+}
+
+void gu::deinit_allowlist_service_v1()
+{
+    std::lock_guard<std::mutex> lock(gu_allowlist_service_init_mutex);
+    assert(gu_allowlist_service_usage > 0);
+    --gu_allowlist_service_usage;
+    if (gu_allowlist_service_usage == 0) gu_allowlist_service = 0;
+}
